--------------------------------------------------------------------------------
-- Alloy4PV -- Alloy for Process Verification
-- Copyright (C) 2012, 2014  Yoann Laurent (yoann.laurent@lip6.fr)

-- This program is free software: you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation, either version 3 of the
-- License, or (at your option) any later version.

-- This program is distributed in the hope that it will be useful, but
-- WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see
-- <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------

module semantic

open util/ordering[State] as Ord
open util/integer
open util/boolean

open syntax
open organizational

abstract sig Status {}
one sig Running, Finished extends Status {}

sig State {
	heldTokens :  ActivityNode -> one Int,
	offers :  ActivityEdge -> one Int,

	localClock : ExecutableNode -> one Int,
	globalClock : Int,

	running : Status	
}

// getters
pred State.hasOffers[edge:ActivityEdge] { this.getOffers[edge] > 0 }
pred State.hasTokens[node:ActivityNode]  { this.getTokens[node] > 0 }
fun State.getTokens[node:ActivityNode] : Int { this.heldTokens[node] }
fun State.getOffers  [edge:ActivityEdge] : Int { this.offers[edge] }

// setters
pred State.setTokens[node:ActivityNode, i : Int]  { this.heldTokens[node] =  i}
pred State.setOffers  [edge:ActivityEdge, i : Int]  { this.offers[edge] = i}

// time
fun State.getLocalClock[node:ExecutableNode] : Int { this.localClock[node]}
pred State.setLocalClock[node:ExecutableNode, i : Int] { this.localClock[node] = i}

pred setExecuting[s,s':State] { s'.running = Running }

// block Edge and Node 
pred blockAll[s,s':State] { blockAllNode[s,s',ActivityNode] and blockAllEdge[s,s',ActivityEdge]}
pred blockAllEdge[s,s':State,edge: set ActivityEdge] { all e : edge | blockEdge[s,s',e] }
pred blockAllNode[s,s':State,node: set ActivityNode] { all n : node | blockNode[s,s',n] }
pred blockEdge[s,s':State, edge:ActivityEdge] {s.getOffers[edge] = s'.getOffers[edge]}
pred blockNode[s,s':State, node:ActivityNode] { s.getTokens[node] = s'.getTokens[node] }

/************************************************************************
Transition Systems
************************************************************************/
pred Init[ n:ActivityNode->Int, e:ActivityEdge->Int]
{
	first.heldTokens    = (ActivityNode -> 0)  ++ n
	first.offers = (ActivityEdge -> 0) ++ e
	first.running = Running

	first.globalClock = 0
	first.localClock = (ExecutableNode -> 0)
}

fact traces {
	all s: State - last | let s' = s.next | {
		s.running = Running implies {
			let canStart = some n:ActivityNode-Pin | s.isReadyToStart[n]  |
			let canFinish = some n:ActivityNode-Pin | s.isReadyToFinish[n] | {
				// if both finish and nodes, either start or finish on a node
				(canStart and canFinish) implies { startNode[s,s'] or finishNode[s,s'] or incTime[s,s'] }
				// finish the execution of one node
				else (canFinish) implies { finishNode[s,s'] or incTime[s,s']}
				// start the execution of one node
				else (canStart) implies { startNode[s,s'] or incTime[s,s']}
				// nothing can be done, termination by stuttering the last state
				else  { endLoop[s,s'] }
			}
		} else {
			// stuttering the last state
			endLoop[s,s']
		}
	}
}

pred startNode[s,s':State] {
	(one n:(ActivityNode-Pin) | s.isReadyToStart[n] and StepStartNode[s,s',n] and blockTime[s,s'])
}

pred finishNode[s,s':State] {
	(one n:(ActivityNode-Pin) | s.isReadyToFinish[n] and StepFinishNode[s,s',n] and blockTime[s,s',n]
										and s.getLocalClock[n].gte[Timing[n]] ) 	
}

pred endLoop[s,s':State] {
	blockAll[s,s']	
	s'.running = Finished 
	s'.localClock = s.localClock
	s'.globalClock = s.globalClock
}

/************************************************************************
Timing
************************************************************************/

pred incTime[s,s':State] {
		setExecuting[s,s']
	blockAll[s,s']
	some en:ExecutableNode | s.hasTokens[en] and not s.getLocalClock[en].gte[Timing[en]]
	// update running localLock
	all n:ExecutableNode | s.hasTokens[n] implies {
		// incremental localClock
		s'.setLocalClock[n,s.getLocalClock[n].add[1]]
	} else {
		s'.setLocalClock[n,s.getLocalClock[n]]
	}
	// update globalClock
	s'.globalClock = s.globalClock.add[1]
}
pred blockTime[s,s':State] {
	all n:ExecutableNode | s'.setLocalClock[ n,s.getLocalClock[n] ]
	s'.globalClock = s.globalClock
}
pred blockTime[s,s':State,r:ActivityNode] {
	r in ExecutableNode implies {
		all n:ExecutableNode-r | s'.setLocalClock[ n,s.getLocalClock[n] ]
		s'.setLocalClock[ r, 0 ]
		s'.globalClock = s.globalClock
	} else {
		blockTime[s,s']
	}

}

/************************************************************************
Step Finish
************************************************************************/
pred State.isReadyToFinish[node:ActivityNode] {
	// the node is already executing
	this.hasTokens[node] 
	and
	// the node owns outgoing edge
	some node.outgoing
}

pred StepFinishNode[s , s' : State, node:ActivityNode] {
	// set the status to executing
	setExecuting[s,s']

	// Action
	node in Action implies {
		s'.setTokens[node, 0]
		// add offers to both outgoing edge and outgoing edge of output
		// the number of added offers corresponds to the number of tokens on the node
		all edge : (node.outgoing+node.output.outgoing) | {
			s'.setOffers[edge, s.getOffers[edge].add[s.getTokens[edge.source]]]	 
		}
		// reset input and output pin
		all pin : (node.output+node.input) | {
			s'.setTokens[pin, 0]
		}
		blockAllEdge[s,s',(ActivityEdge-(node.outgoing+node.output.outgoing))]	
		blockAllNode[s,s',(ActivityNode-(node+node.input+node.output))]						
	} 

	// DecisionNode
	else node in DecisionNode implies {
		s'.setTokens[node, 0]
		// offer tokens on only one outgoing edge
		one edge : node.outgoing | {
			s'.setOffers[edge, s.getOffers[edge].add[s.getTokens[edge.source]]]	 
			blockAllEdge[s,s',(ActivityEdge-edge)]
		}
		blockAllNode[s,s',(ActivityNode-node)]
	}		

	// FlowFinalNode
	else node in FlowFinalNode implies {
		s'.setTokens[node, 0]
		blockAllEdge[s,s',(ActivityEdge)]
		blockAllNode[s,s',(ActivityNode-node)]
	}		
	 
	// (default) ActivityNode
	else node in ActivityNode implies {	
		s'.setTokens[node, 0]
		all edge : node.outgoing | {
			s'.setOffers[edge, s.getOffers[edge].add[s.getTokens[edge.source]]]	 
		}
		blockAllEdge[s,s',(ActivityEdge-node.outgoing)]
		blockAllNode[s,s',(ActivityNode-node)]
	}
}

/************************************************************************
Step Start
************************************************************************/
pred State.isReadyToStart(node:ActivityNode) {
	// the node is not already executing
	not this.hasTokens[node]
	and	
	// the node owns incoming edge
	some node.incoming
	and
	// if the node is a MergeNode
	// only 1 incoming edge with offers is required
	(node in MergeNode implies { 
		some n:node.incoming | this.hasOffers[n]
	}
	// if the node is a Action
	// check if all the incoming edge have offers
	// check if all the input pin have enough offers depending of their multiplicity
	else node in Action implies {
		all n:node.incoming | this.hasOffers[n]
		and
		all n:node.input.incoming | this.getOffers[n] >=n.target.multiplicityElement.lower
	} 
	// (default) If the node is a ActivityNode
	// check if all the incoming edge have offers
	else node in ActivityNode implies {	
		all n:node.incoming | this.hasOffers[n]
	})	
}

/* receiveOffer() */
pred StepStartNode[s , s' : State, node : ActivityNode] {
	// set the status to executing
	setExecuting[s,s']

	// Action
	node in Action implies {	
		s'.setTokens[node, 1]
		// set the number of consumed tokens in the input pin
		// trying to consume all offered tokens	
		all iPin : node.input | {
			iPin.multiplicityElement.upper >= s.getOffers[iPin.incoming]
			implies
				// consume all
				s'.getTokens[iPin] = s.getOffers[iPin.incoming]	
			else
				// consume up to upper
				s'.getTokens[iPin] = iPin.multiplicityElement.upper 
		}
		// set the number of tokens in the output pin
		// nondeterministic between lower and upper multiplicity
		all oPin : node.output | {
			s'.getTokens[oPin] >= oPin.multiplicityElement.lower
			and	
			s'.getTokens[oPin] <= oPin.multiplicityElement.upper
		}
		// remove offers from both incoming edge and incoming edge of input
		// the number of removed offers corresponds to the new number of tokens on the node
		all edge : (node.incoming+node.input.incoming)  | {
			s'.setOffers[edge, s.getOffers[edge].sub[s'.getTokens[edge.target]]]	 
		}
		blockAllEdge[s,s',ActivityEdge-(node.incoming+node.input.incoming) ]
		blockAllNode[s,s',ActivityNode-(node+node.input+node.output)]
	}

	// MergeNode
	else node in MergeNode implies {	
		s'.setTokens[node, 1]
		// choose one incoming edge
		one edge : node.incoming | {  
			// the choosed edge need to have offers
			s.hasOffers[edge]
			// remove offers from the edge
			s'.setOffers[edge, s.getOffers[edge].sub[s'.getTokens[edge.target]]]	 
			blockAllEdge[s,s',(ActivityEdge-edge)]
		}
		blockAllNode[s,s',(ActivityNode-node)]
	}	

	// ActivityFinalNode
	else node in ActivityFinalNode implies {	
		s'.setTokens[node, 1]
		all edge : node.incoming | {
			s'.setOffers[edge, s.getOffers[edge].sub[s'.getTokens[edge.target]]]	 
		}
		blockAllEdge[s,s',(ActivityEdge-node.incoming)]
		blockAllNode[s,s',(ActivityNode-node)]
		s'.next.running = Finished 
	}

	// (default) ActivityNode
	else  node in ActivityNode implies {
		s'.setTokens[node, 1]
		all edge : node.incoming | {
			s'.setOffers[edge, s.getOffers[edge].sub[s'.getTokens[edge.target]]]	 
		}
		blockAllEdge[s,s',(ActivityEdge-node.incoming)]
		blockAllNode[s,s',(ActivityNode-node)]
	}
}


/************************************************************************
Fairness
************************************************************************/
// (1)
pred fairness1 {
	some s:State | s.running = Finished
}
// (2)
pred fairness2 {
	all decision:DecisionNode | fairDecision[decision]
}
pred fairDecision[node:DecisionNode] {
	all edge:node.outgoing | 
		#{getNumberOfEdgeActivation[edge]} <= 2
}
fun getNumberOfEdgeActivation[edge:ActivityEdge] : State {
	{ s : State | s.getOffers[edge] < s.next.getOffers[edge]  }
}
